@@ -1,6 +1,6 @@
 /* 
    SSL interface definitions internal to neon.
-   Copyright (C) 2003-2005, 2008, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2003-2021, Joe Orton <joe@manyfish.co.uk>
    Copyright (C) 2004, Aleix Conchillo Flaque <aleix@member.fsf.org>
 
    This library is free software; you can redistribute it and/or
@@ -40,15 +40,17 @@ struct ne_ssl_context_s {
     SSL_CTX *ctx;
     SSL_SESSION *sess;
     const char *hostname; /* for SNI */
+    int failures; /* bitmask of exposed failure bits. */
 };
 
 typedef SSL *ne_ssl_socket;
 
 /* Create a clicert object from cert DER {der, der_len}, using given
  * RSA_METHOD for the RSA object. */
-ne_ssl_client_cert *ne__ssl_clicert_exkey_import(const unsigned char *der,
-                                                 size_t der_len,
-                                                 const RSA_METHOD *method);
+NE_PRIVATE ne_ssl_client_cert *
+ne__ssl_clicert_exkey_import(const unsigned char *der,
+                             size_t der_len,
+                             const RSA_METHOD *method);
 
 #endif /* HAVE_OPENSSL */
 
@@ -56,8 +58,12 @@ ne_ssl_client_cert *ne__ssl_clicert_exkey_import(const unsigned char *der,
 
 #include <gnutls/gnutls.h>
 
+#ifdef HAVE_GNUTLS_PRIVKEY_IMPORT_EXT
+#include <gnutls/abstract.h>
+#endif
+
 struct ne_ssl_context_s {
-    gnutls_certificate_credentials cred;
+    gnutls_certificate_credentials_t cred;
     int verify; /* non-zero if client cert verification required */
 
     const char *hostname; /* for SNI */
@@ -65,10 +71,10 @@ struct ne_ssl_context_s {
     /* Session cache. */
     union ne_ssl_scache {
         struct {
-            gnutls_datum key, data;
+            gnutls_datum_t key, data;
         } server;
 #if defined(HAVE_GNUTLS_SESSION_GET_DATA2)
-        gnutls_datum client;
+        gnutls_datum_t client;
 #else
         struct {
             char *data;
@@ -76,27 +82,25 @@ struct ne_ssl_context_s {
         } client;
 #endif
     } cache;
-
-#ifdef HAVE_GNUTLS_SIGN_CALLBACK_SET
-    gnutls_sign_func sign_func;
-    void *sign_data;
-#endif
 };
 
-typedef gnutls_session ne_ssl_socket;
+typedef gnutls_session_t ne_ssl_socket;
 
-ne_ssl_client_cert *ne__ssl_clicert_exkey_import(const unsigned char *der,
-                                                 size_t der_len);
+NE_PRIVATE ne_ssl_client_cert *
+ne__ssl_clicert_exkey_import(const unsigned char *der, size_t der_len,
+                             gnutls_privkey_sign_func sign_func, void *userdata);
 
 #endif /* HAVE_GNUTLS */
 
-ne_ssl_socket ne__sock_sslsock(ne_socket *sock);
+#ifdef NE_HAVE_SSL
+NE_PRIVATE ne_ssl_socket ne__sock_sslsock(ne_socket *sock);
 
 /* Process-global initialization of the SSL library; returns non-zero
  * on error. */
-int ne__ssl_init(void);
+NE_PRIVATE int ne__ssl_init(void);
 
 /* Process-global de-initialization of the SSL library. */
-void ne__ssl_exit(void);
+NE_PRIVATE void ne__ssl_exit(void);
+#endif
 
 #endif /* NE_PRIVSSL_H */
