@@ -1,6 +1,6 @@
 /* 
    Tests for property handling
-   Copyright (C) 2002-2006, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2002-2009, Joe Orton <joe@manyfish.co.uk>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -36,8 +36,7 @@
 #include "utils.h"
 
 static const ne_propname p_alpha = {"DAV:", "alpha"},
-    p_beta = {"http://webdav.org/random/namespace", "beta"},
-    p_delta = {NULL, "delta"};
+    p_beta = {"http://webdav.org/random/namespace", "beta"};
 
 /* Tests little except that ne_proppatch() doesn't segfault. */
 static int patch_simple(void)
@@ -53,8 +52,8 @@ static int patch_simple(void)
 		      "HTTP/1.1 200 Goferit\r\n"
 		      "Connection: close\r\n\r\n"));
     ONREQ(ne_proppatch(sess, "/fish", ops));
-    ne_session_destroy(sess);
-    return await_server();
+
+    return destroy_and_wait(sess);
 }
 
 #define RESP207 "HTTP/1.0 207 Stuff\r\n" "Server: foo\r\n\r\n"
@@ -242,13 +241,15 @@ static int tos_endprop(void *userdata, int state,
 static int run_207_response(char *resp, const char *expected)
 {
     ne_buffer *buf = ne_buffer_create();
-    ne_session *sess = ne_session_create("http", "localhost", 7777);
+    ne_session *sess;
     ne_xml_parser *p = ne_xml_create();
     ne_207_parser *p207;
-    ne_request *req = ne_request_create(sess, "PROPFIND", "/foo");
+    ne_request *req;
     ne_uri base = {0};
     struct propctx ctx;
 
+    CALL(session_server(&sess, single_serve_string, resp));
+    req = ne_request_create(sess, "PROPFIND", "/foo");
     ne_fill_server_uri(sess, &base);
     base.path = ne_strdup("/foo");
     p207 = ne_207_create(p, &base, buf);
@@ -263,7 +264,6 @@ static int run_207_response(char *resp, const char *expected)
     ctx.p207 = p207;
     ne_xml_push_handler(p, tos_startprop, tos_cdata, tos_endprop, &ctx);
 
-    CALL(spawn_server(7777, single_serve_string, resp));
 
     ONREQ(ne_request_dispatch(req));
 
@@ -324,7 +324,7 @@ static int two_oh_seven(void)
           "-status={999 French Fries};" 
         },
 
-        /* test multiple propstats in mulitple responses */
+        /* test multiple propstats in multiple responses */
         { MULTI_207(RESP_207("/al/pha", 
                              PSTAT_207(STAT_207("321 Une"))
                              PSTAT_207(STAT_207("432 Deux"))
@@ -411,10 +411,12 @@ static int two_oh_seven(void)
 }
 
 /* Serialize propfind result callbacks into a string */
-static int simple_iterator(void *buf, const ne_propname *name,
+static int simple_iterator(void *vbuf, const ne_propname *name,
                            const char *value, const ne_status *st)
 {
     char code[20];
+    ne_buffer *buf = vbuf;
+
     ne_buffer_concat(buf, "prop:[{", name->nspace, ",", 
                      name->name, "}=", NULL);
     if (value)
@@ -492,6 +494,7 @@ static void pf_destructor(void *userdata, void *private)
 enum pftype { 
     PF_SIMPLE, /* using ne_simple_propfind */
     PF_NAMED,  /* using ne_propfind_named */
+    PF_SP_NAMED, /* using ne_propfind_named w/SHAREPOINT hacks */
     PF_ALLPROP /* using ne_propfind_allprop */
 };
 
@@ -509,7 +512,12 @@ static int run_propfind(const ne_propname *props, char *resp,
     }
     else {
         ne_propfind_handler *hdl;
-        
+
+        if (type == PF_SP_NAMED) {
+            ne_set_session_flag(sess, NE_SESSFLAG_SHAREPOINT, 1);
+            type = PF_NAMED;
+        }
+
         hdl = ne_propfind_create(sess, "/propfind", depth);
 
         ne_propfind_set_private(hdl, pf_creator, pf_destructor,
@@ -590,19 +598,45 @@ static int propfind(void)
           0, PF_SIMPLE},
 
         /* whitespace handling. */
-        { MULTI_207(RESP_207("\r\nhttp://localhost:7777/alpha ",
+        { MULTI_207(RESP_207("\r\nhttp://localhost/alpha ",
                              PSTAT_207(PROPS_207(APROP_207("alpha", "beta"))
                                        "<D:status>\r\nHTTP/1.1 200 OK </D:status>"))),
           "results(/alpha,prop:[{DAV:,alpha}='beta':{200 OK}];)//",
           0, PF_SIMPLE},
         
+        /* attribute handling. */
+        { MULTI_207(RESP_207("\r\nhttp://localhost/alpha ",
+                             PSTAT_207(PROPS_207("<D:alpha>"
+                                                 "<D:foo D:fee='bar' bar='fee'>beta</D:foo></D:alpha>")
+                                       "<D:status>\r\nHTTP/1.1 200 OK </D:status>"))),
+          "results(/alpha,prop:[{DAV:,alpha}='<DAV:foo DAV:fee='bar' bar='fee'>beta</DAV:foo>':{200 OK}];)//",
+          0, PF_SIMPLE},
+        
+
         /* "complex" propfinds. */
 
         { FISHBONE_RESP("hello, world", "212 Well OK"),
           "creator[/foop]//"
           "results(/foop,prop:[{DAV:,fishbone}='hello, world':{212 Well OK}];)//"
           "destructor[/foop]//",
-          0, PF_NAMED }
+          0, PF_NAMED },
+
+        /* 207 with badly encoded URI in href */
+        { MULTI_207(RESP_207("http://example.com/foo€bar", \
+                             PSTAT_207(PROPS_207(APROP_207("fishbone", "hello, world")) \
+                                       STAT_207("209 Good News")))),
+          "creator[/foo%e2%82%acbar]//"
+          "results(/foo%e2%82%acbar,prop:[{DAV:,fishbone}='hello, world':{209 Good News}];)//"
+          "destructor[/foo%e2%82%acbar]//",
+          0, PF_SP_NAMED },
+
+        { MULTI_207(RESP_207("/foo%20bar/€bar", \
+                             PSTAT_207(PROPS_207(APROP_207("fishbone", "hello, world")) \
+                                       STAT_207("209 Good News")))),
+          "creator[/foo%20bar/%e2%82%acbar]//"
+          "results(/foo%20bar/%e2%82%acbar,prop:[{DAV:,fishbone}='hello, world':{209 Good News}];)//"
+          "destructor[/foo%20bar/%e2%82%acbar]//",
+          0, PF_SP_NAMED }
 
     };
     const ne_propname pset1[] = {
@@ -626,12 +660,9 @@ static int unbounded_response(const char *header, const char *repeats)
 {
     ne_session *sess;
     struct infinite i = { header, repeats};
-    int dbg;
 
     CALL(make_session(&sess, serve_infinite, &i));
 
-    dbg = ne_debug_mask;
-
     ONN("unbounded PROPFIND response did not fail",
         ne_simple_propfind(sess, "/", 0, NULL, 
                            dummy_results, NULL) != NE_ERROR);
