@@ -13,6 +13,10 @@ MALLOC_CHECK_=2
 MALLOC_PERTURB_=`expr $RANDOM % 255 2>/dev/null`
 export MALLOC_CHECK_ MALLOC_PERTURB_
 
+export GNUTLS_SYSTEM_PRIORITY_FILE=$PWD/neon.prio
+
+export TEST_QUIET=${TEST_QUIET:-1}
+
 RETVAL=0
 
 for f in $*; do
